#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import testsuite

child = testsuite.spawn('numserv')

child.expect('Waiting for input.')

child.sendline('hundred')
child.expect('100')

child.livepatch('.libs/libhundreds_livepatch1.so')

child.sendline('hundred')
child.expect('200', reject='100')

# Now try to patch a wrong file but revert all patches associated with
# libhundred. If there is a fail on apply and revert-all is specified,
# it shoudn't revert the livepatches. That could imply in unsecure
# code being run.

try:
  child.livepatch('.libs/libnonexistent_livepatch1.so', revert_lib="libhundreds.so.0", sanity=False)
except:
  pass

child.sendline('hundred')
child.expect('200', reject='100')
exit(0)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time

import testsuite
from testsuite import ulptool

errorcode = 0
child = testsuite.spawn('parameters', env=None)

child.expect('Waiting for input.')

child.sendline('')
child.expect('1-2-3-4-5-6-7-8');
child.expect('1.0-2.0-3.0-4.0-5.0-6.0-7.0-8.0-9.0-10.0');

out = child.livepatch('.libs/libparameters_livepatch1.so', capture_tool_output=True)
if out.find("Libpulp not found in target process") == -1:
  errorcode = 1

child.sendline('')
child.expect('1-2-3-4-5-6-7-8', reject='8-7-6-5-4-3-2-1');

child.close(force=True)
exit(errorcode)



#!/usr/bin/env python3

import argparse
import subprocess

def find_offset(file, name):
  nm = subprocess.Popen(['nm', file], stdout=subprocess.PIPE,
                        encoding='utf-8')
  for entry in nm.stdout.readlines():
    split = entry.split(sep=' ')
    symbol = split[2].rstrip('\n')
    if symbol == name:
      return split[0]
  return

# This program takes two arguments, first the path to the input file, then the
# path to the output file. The input file is a live patch description template,
# where every line starting with '#' contains a local (not-exported) variable
# whose address in the target library (__TARGET_OFFSET__), as well as the
# address of its reference in the live patch (__PATCH_OFFSET__) must be
# determined.
parser = argparse.ArgumentParser()
parser.add_argument('ifile')
parser.add_argument('ofile')
args = parser.parse_args()

ifile = open(args.ifile, 'r')
ofile = open(args.ofile, 'w')

# The path to the patch file is always at the first line
patch = ifile.readline()
patch = patch.rstrip('\n')

# The path to the target library is always at the second line,
# which always starts with '@'
target = ifile.readline()
target = target.rstrip('\n')
target = target.rstrip('.json')
target = target.lstrip('@')

# Rewind the input file
ifile.seek(0)

# Iterate over all lines of the input file
for line in ifile:

  # Lines starting with '#' contain local variables
  if line[0] == '#':

    # If there is a % then it is a TLS variable.
    if line[1] == '%':
      split = line.lstrip('#%')
    else:
      split = line.lstrip('#')

    # Parse the line
    split = split.split(':')

    # Get the name of the local variable in the target library
    tname = split[0]

    # Get the name of the local variable reference in the live patch
    pname = split[1]

    # Search for the local variable in the target library
    toff = find_offset(target, tname)

    # Search for the local variable reference in the live patch
    poff = find_offset(patch, pname)

    # Replace offset template patterns with actual offsets
    if toff is not None:
        line = line.replace('__TARGET_OFFSET__', toff)
    if poff is not None:
        line = line.replace('__PATCH_OFFSET__', poff)

  # Write every line back to the output file
  ofile.write(line)

ifile.close()
ofile.close()



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

import testsuite

# Since the deadlock demonstrated by this test case does not occur
# everytime it executes, run it in a loop. The amount of iterations
# hardcoded is arbitrary
errors = 0

child = testsuite.spawn('deadlock', log=None)

child.expect('Waiting for input.')

child.sendline('')
child.expect('hello')

# Applying a live patch to a process entails stopping all of its
# threads, then stealing one of them to jack into the process and call
# libpulp.so's routines that load and apply the live patch. These
# routines are called from the context of a signal-handler, and, as
# such, should not make calls to Asynchronous Signal Unsafe functions.
# However, libpulp calls dlopen, which is AS-Unsafe.
for attempt in range(200):
  print('  Attempt #' + str(attempt))
  try:
    child.livepatch('libblocked_livepatch1.ulp', timeout=10)
  except subprocess.TimeoutExpired:
    print('Deadlock detected.')
    errors = 1
    break
  except subprocess.CalledProcessError:
    pass
  else:
    break

child.close(force=True)
exit(errors)




#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2023 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import re
import os
import testsuite

# Live patch selection variable
child = testsuite.spawn('numserv')

child.expect('Waiting for input.')

child.sendline('dozen')
child.expect('12');

child.sendline('hundred')
child.expect('100');

child.livepatch('.libs/libdozens_livepatch1.so')

child.sendline('dozen')
child.expect('13', reject='12');

child.sendline('hundred')
child.expect('100');

child.disable_livepatching()
child.livepatch('.libs/libhundreds_livepatch1.so', sanity=False)

child.sendline('dozen')
child.expect('13', reject='12');

child.sendline('hundred')
child.expect('100', reject='200');

child.enable_livepatching()
child.livepatch('.libs/libhundreds_livepatch1.so')

child.sendline('hundred')
child.expect('200', reject='100');

child.close(force=True)

#-- Test user name and group id mode.

child = testsuite.spawn('numserv')

child.expect('Waiting for input.')

child.sendline('dozen')
child.expect('12');

child.sendline('hundred')
child.expect('100');

testsuite.childless_disable_livepatching('numserv', os.getgid())
child.livepatch('.libs/libhundreds_livepatch1.so', sanity=False)

child.sendline('dozen')
child.expect('12', reject='13');

child.close(force=True)
exit(0)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import signal
import subprocess

import testsuite

child = testsuite.spawn('asunsafe_conversion')

child.expect('Waiting for signals.')

child.kill(signal.SIGUSR1)
child.expect('hello')

errors = 0
try:
  # Apply the live patch.
  child.livepatch('.libs/libblocked_livepatch1.so', retries=2000)
except subprocess.TimeoutExpired:
  print('Deadlock while live patching - AS-Unsafe conversion not tested')
  # The deadlock test (tests/deadlock) has a far greater chance of
  # detecting deadlocks during the application of live-patches, so
  # return 77 to report that this test case was unable to detect the
  # AS-Unsafe conversion.
  errors = 77
except subprocess.CalledProcessError as err:
  # The trigger tool may fail to apply a live patch when the locks it
  # needs to acquire from glibc are busy. This is expected to happen
  # some times, but not the intent of this test case, so return 77 to
  # report that the AS-Unsafe conversion was not properly detected.
  if err.returncode == 1:
    print('Unable to live patch - AS-Unsafe conversion not tested')
    errors = 77
  # On the other hand, if the trigger tool fails for some unexpected
  # reason, signal it as a hard error (99) to the test harness.
  else:
    print('Unexpected failure while live patching')
    errors = 99
else:
  # Send signals to the process which will cause the live-patched
  # function to be called from the context of a signal-handler. If the
  # live-patched function became AS-Unsafe, it might deadlock. Since it
  # is not guaranteed to deadlock every time, loop it an arbitrarily
  # chosen number of times.
  for attempt in range(1000):
    child.kill(signal.SIGUSR1)
    try:
      child.expect('olleh')
    except TimeoutError:
      print('AS-Unsafe conversion detected')
      errors = 1
      break

child.close(force=True)
exit(errors)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import testsuite

child = testsuite.spawn('pagecross')

child.expect('Waiting for input.')

child.sendline('')
child.expect('default');

child.livepatch('.libs/libpagecross_livepatch1.so')

child.sendline('')
child.expect('patched', reject='default');

child.close(force=True)
exit(0)



#!/usr/bin/env python3
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import testsuite
import subprocess


childs = [testsuite.spawn('manyprocesses'),
          testsuite.spawn('manyprocesses'),
          testsuite.spawn('manyprocesses')]

for child in childs:
    child.expect('Waiting for input.')

    child.sendline('')
    child.expect('1-2-3-4-5-6-7-8');
    child.expect('1.0-2.0-3.0-4.0-5.0-6.0-7.0-8.0-9.0-10.0');

# Try to livepatch everything into the manyprocesses.
output = testsuite.childless_livepatch(wildcard='.libs/*_livepatch1.so',
                                       verbose=True,
                                       revert_lib='libmanyprocesses.so.0',
                                       capture_output=True)

# Check if the patched counter is correct.
if output.find("Processes patched: 3, Skipped: 0, Failed: 0") == -1:
    exit(1)


for child in childs:
    child.sendline('')
    child.expect('8-7-6-5-4-3-2-1', reject='1-2-3-4-5-6-7-8');
    child.expect('10.0-9.0-8.0-7.0-6.0-5.0-4.0-3.0-2.0-1.0',
                 reject='1.0-2.0-3.0-4.0-5.0-6.0-7.0-8.0-9.0-10.0');

testsuite.childless_livepatch(wildcard=None, verbose=True, revert_lib='libmanyprocesses.so.0')

for child in childs:
    child.sendline('')
    child.expect('1-2-3-4-5-6-7-8', reject='8-7-6-5-4-3-2-1');
    child.expect('1.0-2.0-3.0-4.0-5.0-6.0-7.0-8.0-9.0-10.0',
                 reject='10.0-9.0-8.0-7.0-6.0-5.0-4.0-3.0-2.0-1.0');

    child.close(force=True)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2023 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

# Test `ulp patches` command.

import testsuite
import subprocess
import sys
import os

# Check if the ULP tool crashes if -p is passed to unexisting process.
command = [testsuite.ulptool, "patches", "-p", "99999"]
try:
    tool = subprocess.run(command, timeout=10, stderr=subprocess.STDOUT)
except:
    exit(1)

# It should actually return an error because we explicitly trieed to retrieve
# information of a non-running process.
if tool.returncode == 0:
    exit(1)

exit(0)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2020-2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time

import testsuite
from testsuite import ulptool

parent = testsuite.spawn('./terminal ./loop', script=False)

# Parent signal readiness first
parent.expect('Parent ready')

# Read the pid of the child process, which wrote it to stdout
parent.expect('Child ready')
parent.expect('[0-9]+')
pid = parent.after
if not type(pid) == str:
  raise TypeError
if not pid.isnumeric():
  raise ValueError

# If live patching SIGSTOPs the target process, the parent detects it
# with waitpid, but only most of the time. If the trigger and check
# tools also send SIGCONT quickly after SIGSTOP, the detection doesn't
# always work, so try it an arbitrarily large number of times.
errors = 0
for i in range(32):

  # Use the Check tool to attach and detach to the child process. Notice
  # that, unlike other test cases, this child process has not been
  # started by the testing framework. Rather, it has been forked from
  # its parent (terminal.c), thus, child.is_patch_applied and
  # child.livepatch cannot be used.
  ret = subprocess.run([ulptool, '-q', '-p', pid, "check",
      '.libs/libblocked_livepatch1.so'])
  # The Check tool returns 0 when the given live patch has not been
  # applied, which is the expected output here. Anything else is treated
  # as a hard error (99), because detecting errors in the Check tool
  # itself is not the intent of this test case.
  if ret.returncode:
    print('Unexpected error while using ulp tools')
    errors = 99
    break
  else:
    print('Check OK')

  # Wait a while so that the parent process has a chance to detect
  # SIGSTOP/SIGCONT and exit normally with exit status 7, a magic
  # number. Other codes are unexpected and treated as hard error (99).
  time.sleep(0.1)
  if parent.isalive() == False:
    if parent.exitstatus == 7:
      print('Parent detected SIGSTOP/SIGCONT and exited')
    else:
      print('Unknown cause for parent termination')
      errors = 99
    break

# Kill the parent process
parent.close(force=True)
if parent.exitstatus == 7:
  if errors == 0:
    errors = 1

# Guarantee the child process gets killed
try:
  os.kill(int(pid), signal.SIGKILL)
except:
  print('Failed to kill forked process. Already dead?')

exit(errors)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

import testsuite

child = testsuite.spawn('memory_protection')

child.expect('Waiting for input.')

# Every time a newline is sent, the test program touchs code memory
child.sendline('')
child.expect('pristine');

child.livepatch('.libs/libaddress_livepatch1.so')

# Try to touch code memory after live patching
child.sendline('')
try:
  child.expect('patched', reject='pristine');
except EOFError:
  # Diagnose the error
  print('Touching code after live patch failed.')
  if child.isalive() == False:
    print('The test program is dead')
  else:
    child.close(force=True)
  # Let the exception propagate, so that the stack trace gets printed
  raise

child.close(force=True)
exit(0)



#!/usr/bin/env python3

#   libpulp - User-space Livepatching Library
#
#   Copyright (C) 2021 SUSE Software Solutions GmbH
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

# From standard library
import os
import pathlib
import re
import signal
import subprocess
import sys
import time

# Third-party libraries
import pexpect
import psutil

# Test case name as provided by automake tests
testname = os.path.splitext(sys.argv[0])
testname = os.path.basename(testname[0])

# Libpulp definitions
builddir = os.getcwd()
ulptool = builddir + '/../tools/ulp'

# Check if certain library is livepatchable.
def is_library_livepatchable(library):
  command = [ulptool, "livepatchable", library]

  try:
    tool = subprocess.run(command, timeout=10, stderr=subprocess.STDOUT)
  except subprocess.TimeoutExpired:
    print('ulp tool deadlock');
    return False

  if tool.returncode == 0:
    return True
  return False

# Wrapper around pexpect.spawn that automatically sets userspace livepatching
# requirements, such as LD_PRELOAD'ing libpulp.so, as well as extends its
# functionality with live patching operations.
#
# Using this testing framework requires few steps: importing this module,
# starting a live patchable process, optionally testing its default behavior,
# applying a live patch, and testing the patched behavior, for instance:
#
#   # Import the testing framework:
#   import testsuite
#
#   # Start the target process:
#   child = testsuite.spawn('testcase')
#
#   # Check default behavior:
#   child.send('ping')
#   child.expect('pong')
#
#   # Apply a live patch
#   child.livepatch('metadata')
#
#   # Check patched behavior:
#   child.send('ping')
#   child.expect('pang')
#
# The 'livepatch' and 'expect' methods are the main interfaces to the testing
# framework. Simply calling them is usually enough, because they check for
# error conditions on their own. However, if more control over the outcomes is
# needed, they raise exceptions which can be caught and dealt with. See the
# existing test cases for examples.
class spawn(pexpect.spawn):

  # Spawn a live patchable process. Similar to pexpect.spawn, this class
  # constructor starts and controls a child process. Unlike pexpect.spawn,
  # LD_PRELOAD=libpulp.so is automatically provided, unless 'env' is passed as
  # an argument. The arguments: 'timeout', 'env' and 'encoding' are forwarded
  # to pexpect.spawn; whereas log is assigned to pexpect's logfile_read
  # attribute. Finally, 'verbose' controls whether the live patching methods
  # print messages to stdout. By default, all messages are printed, so that the
  # test suite logs contain more information for debugging.
  def __init__(self, testname, timeout=10, env=Ellipsis, log=sys.stdout,
               encoding='utf-8', verbose=True, script=True):

    # If testname is a relative path, i.e. not starting with slash, prepend
    # dot-slash to enable command-line execution.
    if testname[0] != '/':
      testname = './' + testname

    testname = '/usr/bin/sh -c ' + '\''+ testname + '\''

    # if TEST_THROUGH_VALGRIND environment variable is defined, append valgrind
    # call on testname. We actually call valgrind on the `sh` call, and it
    # still catch memory issues while avoiding libpulp problems regarding
    # the program actually being `valgrind` and not the test program.
    try:
        if os.environ['TESTS_THROUGH_VALGRIND'] == '1':
            testname = 'valgrind --leak-check=full ' + testname
    except KeyError:
        pass

    # If env has not been provided, default to LD_PRELOAD'ing libpulp.so.
    if env == Ellipsis:
      env = {'LD_PRELOAD': builddir + '/../lib/.libs/libpulp.so'}

    # Spawn the testcase with pexpect and enable logging.
    super().__init__(testname, timeout=timeout, env=env, encoding=encoding,
                     echo=False)
    self.logfile_read = log
    self.verbose = verbose

  # Print verbose messages when in verbose mode
  def print(self, *args, end=None):
    if self.verbose:
      print(*args, flush=True, end=end)

  # When pexpect.expect is unable to find the expected patterns, it prints
  # comprehensive information about the scanning status, such as the whole
  # child output, as well as less relevant information about the child process.
  # Even though the output is useful, it pollutes the test suite logs
  # unnecessarily. Thus, this class overrides expect() and always adds
  # pexpect.EOF and pexpect.TIMEOUT to the list of expected patterns, so that
  # it can print less verbose, more targeted information, when the original
  # expected patterns and the actual output do not match. If it happens, the
  # overridden method raises exceptions for EOF and Timeout, which have the
  # added benefit of displaying local, shorter, and more friendly stack traces.
  # The 'accept' and 'reject' parameters can be either single strings or lists
  # of strings, such as pexpect.expect expects.
  def expect(self, accept, reject=None):

    # Converted string arguments into lists
    if type(accept) == str:
      accept = [accept]
    if type(reject) == str:
      reject = [reject]
    if reject == None:
      reject = []

    # Valgrind always print the error summary message. Match if the errors are
    # a positive value.
    valgrind_errors = ['ERROR SUMMARY: [1-9]+ errors from',
                       'definitely lost: [1-9]+']

    reject += valgrind_errors
    valgrind_reject_index = len(reject) - len(valgrind_errors)

    # Also add EOF and TIMEOUT to the expected patterns to avoid the verbose
    # output of pexpect.expect when the expected patterns are not found
    patterns = [pexpect.EOF, pexpect.TIMEOUT] + accept + reject

    # Actually look for the expected patterns
    index = super().expect(patterns)

    # Handle EOF and TIMEOUT, then raise the corresponding exception
    if index == 0 or index == 1:
      self.print('error: expected output not found.')
      self.print('expected:', accept, reject)
      self.print('observed: %r' %(self.buffer[-self.str_last_chars:]))
    if index == 0:
      raise EOFError
    if index == 1:
      raise TimeoutError
    # Account for the prepended items in the patterns list
    index = index - 2

    # If the matching pattern belongs to the accepted list, return its index
    if index < len(accept):
      self.print('Accept pattern found:', accept[index])
      return index

    # Otherwise, the matching pattern belongs to the rejected list
    index = index - len(accept)
    if index >= valgrind_reject_index:
        self.print('Valgrind error detected.')
        raise MemoryError
    else:
        self.print('Reject pattern found:', reject[index])
        raise ValueError

  # Verify sanity of arguments to ulp tools
  def sanity(self, filename=Ellipsis, pid=Ellipsis):

    # Check if the process with PID exists
    if not pid == Ellipsis:
      if not psutil.pid_exists(pid):
        raise ValueError('Process ' + str(pid) + ' not found')

    # Check if the live patch file exists
    if not filename == Ellipsis and filename is not None:
      file = pathlib.Path(filename)
      if not file.is_file():
        raise FileNotFoundError('File ' + filename + ' not found')

  # Dump a compiled metadata file (.ulp) provided in metadata
  def show_metadata(self, metadata):
    # See if there is a .dsc file. On true, print it as well.
    suffix = metadata.rfind(".ulp")
    dsc = metadata[:suffix] + ".dsc"
    if os.path.isfile(dsc):
      self.print(dsc + " content:")
      dscf = open(dsc, 'r')
      content = dscf.read()
      dscf.close()
      self.print(content)

    # Now print the .ulp file
    command = [ulptool, "dump", metadata]
    try:
      self.print(metadata + ' content:')
      tool = subprocess.run(command, timeout=10)
    except subprocess.TimeoutExpired:
      self.print('ulp dump timed out.')
      raise


  # Apply a live patch to the spawned process. The path to the live patch
  # metadata must be passed through 'filename'. The remaining parameters, which
  # are optional, are the same that the Trigger tool provides (see its --help
  # output for more information).
  def livepatch(self, filename=None, timeout=10, retries=1,
                verbose=True, quiet=False, revert=False, revert_lib=None,
                sanity=True, prefix=None, capture_tool_output=False):

    # Check sanity of command-line arguments
    if sanity is True:
      self.sanity(pid=self.pid)
      self.sanity(filename=filename)

    # Build command-line from arguments
    command = [ulptool, "trigger", '-p', str(self.pid)]
    if revert is True:
      command.append("--revert")
    if revert_lib is not None:
      command.append("--revert-all")
      command.append(revert_lib)
    if filename is not None:
      command.append(filename)
      self.print('')
      self.show_metadata(filename)
    if verbose:
      command.append('-v')
    if quiet:
      command.append('-q')
    if retries > 1:
      command.append('-r')
      command.append(str(retries))
    if prefix is not None:
      command.append('-R')
      command.append(str(prefix))

    # Apply the live patch and check for common errors
    try:
      self.print('Applying/reverting live patch.')
      if capture_tool_output == True:
          tool = subprocess.run(command, timeout=timeout,
                                stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
      else:
          tool = subprocess.run(command, timeout=timeout)
    except subprocess.TimeoutExpired:
      self.print('Live patching timed out.')
      raise

    self.print("---- Messages in libpulp.so ----")
    self.print(self.get_libpulp_messages())

    # The trigger tool returns 0 on success, so use check_returncode(),
    # which asserts that, and raises CalledProcessError otherwise.
    if sanity is True:
      tool.check_returncode()
    if revert == True:
      self.print('Live patch reverted successfully.')
    else:
      self.print('Live patch applied successfully.')

    if tool.stdout is not None:
      return tool.stdout.decode()
    else:
      return None

  # Check if a live patch is already applied. The path to the live patch
  # metadata must be passed through 'filename'. The remaining parameters, which
  # are optional, are the same that the Checker tool provides (see its --help
  # output for more information).
  def is_patch_applied(self, filename, verbose=True, quiet=False):

    # Check sanity of command-line arguments
    self.sanity(pid=self.pid)
    self.sanity(filename=filename)

    # Build command-line from arguments
    command = [ulptool, '-p', str(self.pid), "check",  filename]
    if verbose:
      command.append('-v')
    if quiet:
      command.append('-q')

    # Apply the live patch and check for common errors
    try:
      self.print('Checking live patch status.')
      tool = subprocess.run(command, timeout=10, stderr=subprocess.STDOUT)
    except subprocess.TimeoutExpired:
      self.print('Live patch status check timed out.')
      raise

    # On success, the checker tool returns either 0 or 1:
    #   1. If the given live patch has already been applied;
    #   0. If it has not.
    if tool.returncode == 0 or tool.returncode == 1:
      self.print('Live patch status check ended successfully. ', end='')
    if tool.returncode == 0:
      self.print('Status is not applied.')
      return False
    if tool.returncode == 1:
      self.print('Status is applied.')
      return True

    # Whereas on failure, the checker tool returns something different,
    # usually -1, so raise CalledProcessError.
    raise subprocess.CalledProcessError

  def is_so_loaded(self, fname_so):
    mapsf = open('/proc/' + str(self.pid) + '/maps', 'r')
    maps = mapsf.read()
    mapsf.close()

    if maps.find(fname_so) == -1:
      return False
    return True

  def get_patches(self, verbose=False):

    # Build command-line from arguments
    command = [ulptool, '-p', str(self.pid), "patches"]
    if verbose:
      command.append('-v')

    try:
      tool = subprocess.run(command, timeout=10, stdout=subprocess.PIPE)
    except subprocess.TimeoutExpired:
      self.print('Live patch status check timed out.')
      raise

    if tool.returncode != 0:
      self.print('ulp tool returned an error code.')
      raise

    msg = str(tool.stdout.decode())
    return msg

  # Get libpulp messages, currently by calling ulp_messages and parsing
  # its stdout.
  def get_libpulp_messages(self):
    self.sanity(pid=self.pid)
    command = [ulptool, 'messages', '-p', str(self.pid)]

    try:
      self.print('Checking libpulp.so messages.')
      tool = subprocess.run(command, timeout=10, stdout=subprocess.PIPE)
    except:
      raise

    msgs = tool.stdout.decode()
    return str(msgs)

  def enable_livepatching(self):
    self.sanity(pid=self.pid)
    command = [ulptool, 'set_patchable', '-p', str(self.pid), 'enable']

    try:
      tool = subprocess.run(command, timeout=10, stdout=subprocess.PIPE)
    except:
      rai2e

    return tool.returncode

  def disable_livepatching(self):
    self.sanity(pid=self.pid)
    command = [ulptool, 'set_patchable', '-p', str(self.pid), 'disable']

    try:
      tool = subprocess.run(command, timeout=10, stdout=subprocess.PIPE)
    except:
      raise

    return tool.returncode


def childless_livepatch(wildcard, timeout=10, retries=1,
            verbose=False, quiet=False, revert_lib=None, capture_output=False):

    # Build command-line from arguments
    command = [ulptool, "trigger"]
    if revert_lib is not None:
      command.append("--revert-all")
      command.append(revert_lib)
    if wildcard is not None:
      command.append(wildcard)
      print('')
    if verbose:
      command.append('-v')
    if quiet:
      command.append('-q')
    if retries > 1:
      command.append('-r')
      command.append(str(retries))

    # Apply the live patch and check for common errors
    try:
      print('Applying/reverting live patch.')
      if capture_output == True:
        tool = subprocess.run(command, timeout=timeout, stdout=subprocess.PIPE)
      else:
        tool = subprocess.run(command, timeout=timeout)
    except subprocess.TimeoutExpired:
      print('Live patching timed out.')
      raise

    # The trigger tool returns 0 on success, so use check_returncode(),
    # which asserts that, and raises CalledProcessError otherwise.
    tool.check_returncode()

    print('Live patch applied/reverted successfully.')
    if tool.stdout is not None:
      return tool.stdout.decode()
    else:
      return None

def childless_disable_livepatching(process_wildcard, userid, timeout=10):

    # Build command-line from arguments
    command = [ulptool, "set_patchable"]
    if process_wildcard is not None:
      command.append("-p")
      command.append(process_wildcard)
    if userid is not None:
      command.append("-u")
      command.append(str(userid))

    command.append('disable')

    # Apply the live patch and check for common errors
    try:
      print('Disabling livepatches.')
      tool = subprocess.run(command, timeout=timeout)
    except subprocess.TimeoutExpired:
      print('Command run timed out')
      raise

    # The trigger tool returns 0 on success, so use check_returncode(),
    # which asserts that, and raises CalledProcessError otherwise.
    tool.check_returncode()



#!/usr/bin/env python3
#
#   This file is part of libpulp.
#
#   libpulp is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   libpulp is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with libpulp.  If not, see <http://www.gnu.org/licenses/>.

# This test stress out the ulp tool when multiple processes are running.

import sys
import testsuite
import subprocess

child = testsuite.spawn('stress')
child.expect("Processes launched")

testsuite.childless_livepatch(wildcard='.libs/libstress_livepatch1.so', verbose=False, timeout=120)

child.expect("Processes finished", reject=['returned non-zero'])
child.close(force=True)
